from numpy._core.records import (
    array,
    find_duplicate,
    format_parser,
    fromarrays,
    fromfile,
    fromrecords,
    fromstring,
    recarray,
    record,
)

__all__ = [
    "record",
    "recarray",
    "format_parser",
    "fromarrays",
    "fromrecords",
    "fromstring",
    "fromfile",
    "array",
    "find_duplicate",
]
